@@ -40,14 +40,14 @@ static void processx__init_global_job_handle(void) {
 
   processx__global_job_handle = CreateJobObjectW(&attr, NULL);
   if (processx__global_job_handle == NULL) {
-    PROCESSX_ERROR("Creating global job object", GetLastError());
+    R_THROW_SYSTEM_ERROR("Creating global job object");
   }
 
   if (!SetInformationJobObject(processx__global_job_handle,
                                JobObjectExtendedLimitInformation,
                                &info,
                                sizeof info)) {
-    PROCESSX_ERROR("Setting up global job object", GetLastError());
+    R_THROW_SYSTEM_ERROR("Setting up global job object");
   }
 }
 
@@ -55,7 +55,16 @@ void R_init_processx_win() {
   /* Nothing to do currently */
 }
 
-SEXP processx__killem_all() {
+SEXP processx__unload_cleanup() {
+
+  if (processx__connection_iocp) CloseHandle(processx__connection_iocp);
+  if (processx__iocp_thread) TerminateThread(processx__iocp_thread, 0);
+  if (processx__thread_start) CloseHandle(processx__thread_start);
+  if (processx__thread_done) CloseHandle(processx__thread_done);
+
+  processx__connection_iocp = processx__iocp_thread =
+    processx__thread_start = processx__thread_done = NULL;
+
   if (processx__global_job_handle) {
     TerminateJobObject(processx__global_job_handle, 1);
     CloseHandle(processx__global_job_handle);
@@ -90,7 +99,7 @@ int processx__utf8_to_utf16_alloc(const char* s, WCHAR** ws_ptr) {
     /* cchWideChar =    */ ws_len);
 
   if (r != ws_len) {
-    error("processx error interpreting UTF8 command or arguments");
+    R_THROW_ERROR("processx error interpreting UTF8 command or arguments: '%s'", s);
   }
 
   *ws_ptr = ws;
@@ -310,10 +319,10 @@ int env_strncmp(const wchar_t* a, int na, const wchar_t* b) {
   B = alloca((nb+1) * sizeof(wchar_t));
 
   r = LCMapStringW(LOCALE_INVARIANT, LCMAP_UPPERCASE, a, na, A, na);
-  if (!r) PROCESSX_ERROR("make environment", GetLastError());
+  if (!r) R_THROW_SYSTEM_ERROR("make environment for process");
   A[na] = L'\0';
   r = LCMapStringW(LOCALE_INVARIANT, LCMAP_UPPERCASE, b, nb, B, nb);
-  if (!r) PROCESSX_ERROR("make environment", GetLastError());
+  if (!r) R_THROW_SYSTEM_ERROR("make environment for process");
   B[nb] = L'\0';
 
   while (1) {
@@ -377,7 +386,8 @@ static int processx__add_tree_id_env(const char *ctree_id, WCHAR **dst_ptr) {
   return 0;
 }
 
-static int processx__make_program_env(SEXP env_block, const char *tree_id, WCHAR** dst_ptr) {
+static int processx__make_program_env(SEXP env_block, const char *tree_id,
+                                      WCHAR** dst_ptr, const char *cname) {
   WCHAR* dst;
   WCHAR* ptr;
   size_t env_len = 0;
@@ -393,7 +403,7 @@ static int processx__make_program_env(SEXP env_block, const char *tree_id, WCHAR
 
   /* first pass: determine size in UTF-16 */
   for (j = 0; j < num; j++) {
-    const char *env = CHAR(STRING_ELT(env_block, 0));
+    const char *env = CHAR(STRING_ELT(env_block, j));
     if (strchr(env, '=')) {
       len = MultiByteToWideChar(CP_UTF8,
                                 0,
@@ -422,7 +432,7 @@ static int processx__make_program_env(SEXP env_block, const char *tree_id, WCHAR
   ptr = dst_copy;
   ptr_copy = env_copy;
   for (j = 0; j < num; j++) {
-    const char *env = CHAR(STRING_ELT(env_block, 0));
+    const char *env = CHAR(STRING_ELT(env_block, j));
     if (strchr(env, '=')) {
       len = MultiByteToWideChar(CP_UTF8,
                                 0,
@@ -502,7 +512,8 @@ static int processx__make_program_env(SEXP env_block, const char *tree_id, WCHAR
                                            ptr,
                                            (int) (env_len - (ptr - dst)));
         if (var_size != len-1) { /* race condition? */
-          PROCESSX_ERROR("GetEnvironmentVariableW", GetLastError());
+          R_THROW_SYSTEM_ERROR("GetEnvironmentVariableW for process '%s'",
+                               cname);
         }
       }
       i++;
@@ -802,29 +813,6 @@ static WCHAR* processx__search_path(const WCHAR *file,
   return result;
 }
 
-void processx__error(const char *message, DWORD errorcode,
-		     const char *file, int line) {
-  LPVOID lpMsgBuf;
-  char *msg;
-
-  FormatMessage(
-    FORMAT_MESSAGE_ALLOCATE_BUFFER |
-    FORMAT_MESSAGE_FROM_SYSTEM |
-    FORMAT_MESSAGE_IGNORE_INSERTS,
-    NULL,
-    errorcode,
-    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-    (LPTSTR) &lpMsgBuf,
-    0, NULL );
-
-  msg = R_alloc(1, strlen(lpMsgBuf) + 1);
-  strcpy(msg, lpMsgBuf);
-  LocalFree(lpMsgBuf);
-
-  error("processx error, %s: #%d %s at '%s:%d'", message,
-	(int) errorcode, msg, file, line);
-}
-
 void processx__collect_exit_status(SEXP status, DWORD exitcode);
 
 DWORD processx__terminate(processx_handle_t *handle, SEXP status) {
@@ -850,6 +838,7 @@ void processx__finalizer(SEXP status) {
   }
 
   if (handle->hProcess) CloseHandle(handle->hProcess);
+  handle->hProcess = NULL;
   R_ClearExternalPtr(status);
   processx__handle_destroy(handle);
 }
@@ -859,7 +848,7 @@ SEXP processx__make_handle(SEXP private, int cleanup) {
   SEXP result;
 
   handle = (processx_handle_t*) malloc(sizeof(processx_handle_t));
-  if (!handle) { error("Out of memory"); }
+  if (!handle) { R_THROW_ERROR("Out of memory when creating subprocess"); }
   memset(handle, 0, sizeof(processx_handle_t));
 
   result = PROTECT(R_MakeExternalPtr(handle, private, R_NilValue));
@@ -876,13 +865,13 @@ void processx__handle_destroy(processx_handle_t *handle) {
   free(handle);
 }
 
-SEXP processx_exec(SEXP command, SEXP args,
-		   SEXP std_in, SEXP std_out, SEXP std_err,
-		   SEXP connections, SEXP env,
-		   SEXP windows_verbatim_args, SEXP windows_hide,
-		   SEXP private, SEXP cleanup, SEXP wd, SEXP encoding,
-		   SEXP tree_id) {
+SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
+                   SEXP std_err, SEXP pty, SEXP pty_options,
+		   SEXP connections, SEXP env, SEXP windows_verbatim_args,
+                   SEXP windows_hide, SEXP private, SEXP cleanup,
+                   SEXP wd, SEXP encoding, SEXP tree_id) {
 
+  const char *ccommand = CHAR(STRING_ELT(command, 0));
   const char *cstd_in = isNull(std_in) ? 0 : CHAR(STRING_ELT(std_in, 0));
   const char *cstd_out = isNull(std_out) ? 0 : CHAR(STRING_ELT(std_out, 0));
   const char *cstd_err = isNull(std_err) ? 0 : CHAR(STRING_ELT(std_err, 0));
@@ -912,26 +901,29 @@ SEXP processx_exec(SEXP command, SEXP args,
 
   err = processx__utf8_to_utf16_alloc(CHAR(STRING_ELT(command, 0)),
 				      &application);
-  if (err) { PROCESSX_ERROR("utf8 -> utf16 conversion", err); }
+  if (err) {
+    R_THROW_SYSTEM_ERROR_CODE(err, "utf8 -> utf16 conversion '%s'",
+                              ccommand);
+  }
 
   err = processx__make_program_args(
       args,
       options.windows_verbatim_args,
       &arguments);
-  if (err) { PROCESSX_ERROR("making program args", err); }
+  if (err) { R_THROW_SYSTEM_ERROR_CODE(err, "making program args '%s'", ccommand); }
 
   if (isNull(env)) {
     err = processx__add_tree_id_env(ctree_id, &cenv);
   } else {
-    err = processx__make_program_env(env, ctree_id, &cenv);
+    err = processx__make_program_env(env, ctree_id, &cenv, ccommand);
   }
-  if (err) PROCESSX_ERROR("making environment", err);
+  if (err) R_THROW_SYSTEM_ERROR_CODE(err, "making environment '%s'", ccommand);
 
   if (ccwd) {
     /* Explicit cwd */
     err = processx__utf8_to_utf16_alloc(ccwd, &cwd);
     if (err) {
-      PROCESSX_ERROR("convert current directory encoding", GetLastError());
+      R_THROW_SYSTEM_ERROR("convert current directory encoding '%s'", ccommand);
     }
 
   } else {
@@ -940,14 +932,14 @@ SEXP processx_exec(SEXP command, SEXP args,
 
     cwd_len = GetCurrentDirectoryW(0, NULL);
     if (!cwd_len) {
-      PROCESSX_ERROR("get current directory length", GetLastError());
+      R_THROW_SYSTEM_ERROR("get current directory length '%s'", ccommand);
     }
 
     cwd = (WCHAR*) R_alloc(cwd_len, sizeof(WCHAR));
 
     r = GetCurrentDirectoryW(cwd_len, cwd);
     if (r == 0 || r >= cwd_len) {
-      PROCESSX_ERROR("get current directory", GetLastError());
+      R_THROW_SYSTEM_ERROR("get current directory '%s'", ccommand);
     }
   }
 
@@ -957,14 +949,14 @@ SEXP processx_exec(SEXP command, SEXP args,
 
     path_len = GetEnvironmentVariableW(L"PATH", NULL, 0);
     if (!path_len) {
-      PROCESSX_ERROR("get env var length", GetLastError());
+      R_THROW_SYSTEM_ERROR("get env var length '%s'", ccommand);
     }
 
     path = (WCHAR*) R_alloc(path_len, sizeof(WCHAR));
 
     r = GetEnvironmentVariableW(L"PATH", path, path_len);
     if (r == 0 || r >= path_len) {
-      PROCESSX_ERROR("get env var", GetLastError());
+      R_THROW_SYSTEM_ERROR("get PATH env var for '%s'", ccommand);
     }
   }
 
@@ -980,15 +972,26 @@ SEXP processx_exec(SEXP command, SEXP args,
   err = processx__stdio_create(handle, extra_connections, num_connections,
 			       cstd_in, cstd_out, cstd_err,
 			       &handle->child_stdio_buffer, private,
-			       cencoding);
-  if (err) { PROCESSX_ERROR("setup stdio", err); }
+			       cencoding, ccommand);
+  if (err) { R_THROW_SYSTEM_ERROR_CODE(err, "setup stdio for '%s'", ccommand); }
 
   application_path = processx__search_path(application, cwd, path);
+
+  /* If a UNC Path, then we try to flip the forward slashes, if any.
+   * It is apparently enough to flip the first two slashes, the rest
+   * are not important. */
+  if (! application_path && wcslen(path) >= 2 &&
+    application[0] == L'/' && application[1] == L'/') {
+    application[0] = L'\\';
+    application[1] = L'\\';
+    application_path = processx__search_path(application, cwd, path);
+  }
+
   if (!application_path) {
     R_ClearExternalPtr(result);
     processx__stdio_destroy(handle->child_stdio_buffer);
     free(handle);
-    error("Command not found");
+    R_THROW_ERROR("Command '%s' not found", ccommand);
   }
 
   startup.cb = sizeof(startup);
@@ -1036,7 +1039,7 @@ SEXP processx_exec(SEXP command, SEXP args,
     /* lpStartupInfo =        */ &startup,
     /* lpProcessInformation = */ &info);
 
-  if (!err) { PROCESSX_ERROR("create process", GetLastError()); }
+  if (!err) { R_THROW_SYSTEM_ERROR("create process '%s'", ccommand); }
 
   handle->hProcess = info.hProcess;
   handle->dwProcessId = info.dwProcessId;
@@ -1064,13 +1067,16 @@ SEXP processx_exec(SEXP command, SEXP args,
        */
       DWORD err = GetLastError();
       if (err != ERROR_ACCESS_DENIED) {
-	PROCESSX_ERROR("Assign to job object", err);
+	R_THROW_SYSTEM_ERROR_CODE(err, "Assign to job object '%s'",
+                                  ccommand);
       }
     }
   }
 
   dwerr = ResumeThread(info.hThread);
-  if (dwerr == (DWORD) -1) PROCESSX_ERROR("resume thread", GetLastError());
+  if (dwerr == (DWORD) -1) {
+    R_THROW_SYSTEM_ERROR("resume thread for '%s'", ccommand);
+  }
   CloseHandle(info.hThread);
 
   processx__stdio_destroy(handle->child_stdio_buffer);
@@ -1086,12 +1092,14 @@ void processx__collect_exit_status(SEXP status, DWORD exitcode) {
   handle->collected = 1;
 }
 
-SEXP processx_wait(SEXP status, SEXP timeout) {
+SEXP processx_wait(SEXP status, SEXP timeout, SEXP name) {
   int ctimeout = INTEGER(timeout)[0], timeleft = ctimeout;
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   processx_handle_t *handle = R_ExternalPtrAddr(status);
   DWORD err, err2, exitcode;
 
-  if (handle->collected) return R_NilValue;
+  if (!handle) return ScalarLogical(1);
+  if (handle->collected) return ScalarLogical(1);
 
   err2 = WAIT_TIMEOUT;
   while (ctimeout < 0 || timeleft > PROCESSX_INTERRUPT_INTERVAL) {
@@ -1107,30 +1115,38 @@ SEXP processx_wait(SEXP status, SEXP timeout) {
   }
 
   if (err2 == WAIT_FAILED) {
-    PROCESSX_ERROR("wait on process", GetLastError());
+    R_THROW_SYSTEM_ERROR("failed to wait on process '%s'", cname);
   } else if (err2 == WAIT_TIMEOUT) {
     return ScalarLogical(FALSE);
   }
 
   /* Collect  */
   err = GetExitCodeProcess(handle->hProcess, &exitcode);
-  if (!err) { PROCESSX_ERROR("get exit code after wait", GetLastError()); }
+  if (!err) {
+    R_THROW_SYSTEM_ERROR("cannot get exit code after wait for '%s'", cname);
+  }
 
   processx__collect_exit_status(status, exitcode);
 
   return ScalarLogical(TRUE);
 }
 
-SEXP processx_is_alive(SEXP status) {
+SEXP processx_is_alive(SEXP status, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   DWORD err, exitcode;
 
+  /* This might happen if it was finalized at the end of the session,
+     even though there are some references to the R object. */
+  if (!handle) return ScalarLogical(0);
+
   if (handle->collected) return ScalarLogical(0);
 
   /* Otherwise try to get exit code */
   err = GetExitCodeProcess(handle->hProcess, &exitcode);
   if (!err) {
-    PROCESSX_ERROR("get exit code to check if alive", GetLastError());
+    R_THROW_SYSTEM_ERROR("failed to get exit code to check if ",
+                         "'%s' is alive", cname);
   }
 
   if (exitcode == STILL_ACTIVE) {
@@ -1141,15 +1157,20 @@ SEXP processx_is_alive(SEXP status) {
   }
 }
 
-SEXP processx_get_exit_status(SEXP status) {
+SEXP processx_get_exit_status(SEXP status, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   DWORD err, exitcode;
 
+  /* This might happen if it was finalized at the end of the session,
+     even though there are some references to the R object. */
+  if (!handle) return R_NilValue;
+
   if (handle->collected) return ScalarInteger(handle->exitcode);
 
   /* Otherwise try to get exit code */
   err = GetExitCodeProcess(handle->hProcess, &exitcode);
-  if (!err) {PROCESSX_ERROR("get exit status", GetLastError()); }
+  if (!err) {R_THROW_SYSTEM_ERROR("get exit status failed for '%s'", cname); }
 
   if (exitcode == STILL_ACTIVE) {
     return R_NilValue;
@@ -1159,10 +1180,12 @@ SEXP processx_get_exit_status(SEXP status) {
   }
 }
 
-SEXP processx_signal(SEXP status, SEXP signal) {
+SEXP processx_signal(SEXP status, SEXP signal, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   DWORD err, exitcode = STILL_ACTIVE;
 
+  if (!handle) return ScalarLogical(0);
   if (handle->collected) return ScalarLogical(0);
 
   switch (INTEGER(signal)[0]) {
@@ -1175,15 +1198,10 @@ SEXP processx_signal(SEXP status, SEXP signal) {
        we are terminating it... */
     err = GetExitCodeProcess(handle->hProcess, &exitcode);
     if (!err) {
-      PROCESSX_ERROR("get exit code after signal", GetLastError());
+      R_THROW_SYSTEM_ERROR("get exit code after signal for '%s'", cname);
     }
 
     if (exitcode == STILL_ACTIVE) {
-
-      /* We only cleanup the tree if the process is still running. */
-      /* TODO: we are not running this for now, until we can properly
-	 work around pid reuse. */
-      /* processx__cleanup_child_tree(handle->dwProcessId); */
       err = processx__terminate(handle, status);
       return ScalarLogical(err != 0);
 
@@ -1197,7 +1215,7 @@ SEXP processx_signal(SEXP status, SEXP signal) {
     /* Health check: is the process still alive? */
     err = GetExitCodeProcess(handle->hProcess, &exitcode);
     if (!err) {
-      PROCESSX_ERROR("get exit code for signal 0", GetLastError());
+      R_THROW_SYSTEM_ERROR("get exit code for signal 0 for '%s'", cname);
     }
 
     if (exitcode == STILL_ACTIVE) {
@@ -1208,24 +1226,26 @@ SEXP processx_signal(SEXP status, SEXP signal) {
   }
 
   default:
-    error("Unsupported signal on this platform");
+    R_THROW_ERROR("Unsupported signal on this platform for '%s'", cname);
     return R_NilValue;
   }
 }
 
-SEXP processx_interrupt(SEXP status) {
-  error("Internal processx error, `processx_interrupt()` should not be called");
+SEXP processx_interrupt(SEXP status, SEXP name) {
+  R_THROW_ERROR("Internal processx error, `processx_interrupt()` should not be called");
   return R_NilValue;
 }
 
-SEXP processx_kill(SEXP status, SEXP grace) {
-  return processx_signal(status, ScalarInteger(9));
+SEXP processx_kill(SEXP status, SEXP grace, SEXP name) {
+  return processx_signal(status, ScalarInteger(9), name);
 }
 
 SEXP processx_get_pid(SEXP status) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
 
-  if (!handle) { error("Internal processx error, handle already removed"); }
+  /* This might happen if it was finalized at the end of the session,
+     even though there are some references to the R object. */
+  if (!handle) return ScalarInteger(NA_INTEGER);
 
   return ScalarInteger(handle->dwProcessId);
 }
@@ -1236,7 +1256,7 @@ SEXP processx__process_exists(SEXP pid) {
   if (proc == NULL) {
     DWORD err = GetLastError();
     if (err == ERROR_INVALID_PARAMETER) return ScalarLogical(0);
-    PROCESSX_ERROR("open process to check if it exists", err);
+    R_THROW_SYSTEM_ERROR_CODE(err, "open process '%d' to check if it exists", cpid);
     return R_NilValue;
   } else {
     /* Maybe just finished, and in that case we still have a valid handle.
@@ -1245,9 +1265,7 @@ SEXP processx__process_exists(SEXP pid) {
     DWORD err = GetExitCodeProcess(proc, &exitcode);
     CloseHandle(proc);
     if (!err) {
-      PROCESSX_ERROR(
-	"get exit code to check if it exists",
-	GetLastError());
+      R_THROW_SYSTEM_ERROR("get exit code to check if process '%d' exists", cpid);
     }
     return ScalarLogical(exitcode == STILL_ACTIVE);
   }
